#!/bin/bash

shellfu import arr
shellfu import pretty

#
# JATS helper functions
#
# JATS is a testing framework specifying format of test suite, test
# result format and few principles.
#
# This libary contains Bash functions to help deal with operations
# around JATS such as parsing test ID's, finding tests and suites,
# etc.
#

#
# Supported version of JATS format
#
JATS__COMPAT=0.0

#
# Colon-separated list of well-known test suite paths
#
JATS__PATH=${JATS__PATH:-/usr/share/jats/suite}

jats__plidparse() {
    #
    # Parse element $1 from plan id $2
    #
    local elem=$1
    local id=$2
    local params
    local nid
    case $id in
        *@*)    params=${id#*@}; id=${id%@$params} ;;
        *)      params=""       ;;
    esac
    nid=${id#jats://}
    case $elem in
        domain|sut_pname|ns)   jats__sidparse "$elem" "$(jats__plidparse suite "$id")" ;;
        suite)      echo "jats://${nid%%//%*}//" ;;
        pln)        echo "${id#*//%}" ;;
        params)     echo "$params" ;;
        "")         warn "usage: jats__plidparse ELEM PLNID"
                    return 2 ;;
        *)          warn "unknown element: $elem"
                    return 2 ;;
    esac
}

jats__sdiscover() {
    #
    # Find suites under dirs $@
    #
    local finddirs=()
    local fmt=ALL
    local tab=$'\t'
    local s_id          # full suite id
    local s_ns          # ^^: namespace
    local s_domain      # ^^: domain
    local s_pname       # ^^: package name
    local s_version     # suite version
    local s_dir         # each suite dir
    local s_fmt         # suite format version
    while true; do case $1 in
        -f) fmt=$2; shift 2 || return 2 ;;
        --) shift; break ;;
        -*) return 2 ;;
        *)  break ;;
    esac done
    finddirs=("$@")
    find "${finddirs[@]}" -type d -path '*/.jats' \
      | sed 's:/\.jats::' \
      | while read -r s_dir; do
            debug -v s_dir
            s_fmt=$(jats__sprop format "$s_dir") || continue
            debug -v s_fmt
            test "$s_fmt" == "$JATS__COMPAT" || continue
            s_domain=$(jats__sprop domain "$s_dir")
            s_ns=$(jats__sprop ns "$s_dir")
            s_pname=$(jats__sprop sut_pname "$s_dir")
            s_version=$(jats__sprop version "$s_dir")
            s_id=$(
                echo -n "jats://$s_domain/$s_ns"
                test -n "$s_pname" && echo -n "/$s_pname"
                echo //
            )
            case $fmt in
                ALL)      echo "$s_id$tab$s_version$tab$s_dir" ;;
                id)       echo "$s_id"      ;;
                path)     echo "$s_dir"     ;;
                version)  echo "$s_version" ;;
            esac
        done
}

jats__sidparse() {
    #
    # Parse element $1 from suite id $2
    #
    local elem=$1   # element of interest
    local id=$2     # id to parse
    local nid       # "naked" id (jats:// prefix dropped)
    local spath     # just ns + sut_pname
    nid=${id#jats://}                               # nid='example.org/our/ns/pkg//'
    #shellcheck disable=SC2001
    spath=$(sed 's://$::' <<<"$nid" | cut -d/ -f2-)   # spath='our/ns/pkg'
    case $elem in
        id)         echo "$id"          ;;      # jats://example.org/our/ns/pkg//
        domain)     echo "${nid%%/*}"   ;;      # example.org
        sut_pname)  echo "${spath##*/}" ;;      # pkg
        ns)         echo "${spath%/*}"  ;;      # our/ns
        "")         warn "usage: jats__sidparse ELEM SUITEID"
                    return 2 ;;
        *)          warn "unknown element: $elem"
                    return 2 ;;
    esac
}

jats__slocate() {
    #
    # Locate suite $1 under known paths
    #
    debug -v JATS__PATH
    local suiteid=$1
    local tab=$'\t'
    local trypath
    local tmatches
    for trypath in ${JATS__PATH//:/ }; do
        debug -v trypath
        test -n "$trypath" || continue
        test -d "$trypath" || continue
        tmatches=$(jattool sfind -f ALL "$trypath" | grep "^$suiteid$tab" )
        case $(grep -c . <<<"$tmatches") in
            1)  cut -d"$tab" -f3 <<<"$tmatches"; return 0 ;;
            *)  continue ;;
        esac
    done
    return 3
}

jats__sprop() {
    #
    # Get property $1 of suite at directory $2
    #
    local pname=$1  # property name
    local sdir=$2   # suite directory
    local pvalue    # property value
    local dvalue    # default property value
    local fpath     # examined file path
    case $sdir in
        .)  fpath=.jats/$pname ;;
        *)  fpath=$sdir/.jats/$pname ;;
    esac
    case $pname in
        domain)         dvalue="jats.example.org" ;;
        ns)             dvalue="anonns" ;;
        sut_nicename)   dvalue=$(jats__sprop sut_pname "$sdir") ;;
        sut_pname)      dvalue="anonsut" ;;
        sut_url)        dvalue="https://example.org/anonsut" ;;
        url)            dvalue="https://$(jats__sprop domain "$sdir")/$(jats__sprop ns "$sdir")" ;;
        format)         : ;;
        version)        dvalue="0.0.0+0._jats_no_version_" ;;
        id)             echo -n "jats://$(jats__sprop domain "$sdir")"
                        echo -n "/$(jats__sprop ns "$sdir")"
                        echo -n "/$(jats__sprop sut_pname "$sdir")"
                        echo -n "//"
                        return 0
                        ;;
        *)              warn "unknown property: $pname"; return 2 ;;
    esac
    pvalue=$(cat "$fpath" 2>/dev/null)
    if test -n "$pvalue"; then
        echo "$pvalue"
        return 0
    elif test "$pname" == "format"; then
        warn "missing mandatory property: $fpath"
        return 3
    else
        echo "$dvalue"
    fi
}

jats__tdiscover() {
    #
    # Find tests under dirs $@
    #
    local finddirs=()
    local fmt=ALL
    local tab=$'\t'
    local s_dir         # suite root dir
    local t_cand        # possible test path
    local t_apath       # absolute path to test directory
    local t_rpath       # relative path to test directory within $s_dir
    local s_id          # suite's jats: id
    while true; do case $1 in
        -f) fmt=$2; shift 2 || return 2 ;;
        --) shift; break ;;
        -*) return 2 ;;
        *)  break ;;
    esac done
    finddirs=("$@")
    find "${finddirs[@]}" -name test -type f \
      | while read -r t_cand; do
            s_dir=$(__jats__straverse "$t_cand") || continue
            t_apath=$(dirname "$(readlink -f "$t_cand")")
            t_rpath=${t_apath#$s_dir/}
            s_id=$(jats__sdiscover -f id "$s_dir")
            s_ver=$(jats__sdiscover -f version "$s_dir")
            t_id=$s_id$t_rpath
            t_path=$s_dir/$t_rpath
            debug -v s_dir s_id t_apath t_rpath t_id t_path
            case $fmt in
                ALL)     echo "$t_id$tab$s_ver$tab$t_path" ;;
                path)    echo "$t_path" ;;
                id)      echo "$t_id"   ;;
                version) echo "$s_ver"  ;;
            esac
        done
}

jats__tidparse() {
    #
    # Parse element $1 from test id $2
    #
    local elem=$1
    local id=$2
    local params
    local nid
    local spath     # just ns + sut_pname
    case $id in
        *@*)    params=${id#*@}; id=${id%@$params} ;;
        *)      params=""       ;;
    esac
    nid=${id#jats://}
    #shellcheck disable=SC2001
    spath=$(sed 's://.*::' <<<"$nid" | cut -d/ -f2-)
    case $elem in
        domain|sut_pname|ns)   jats__sidparse "$elem" "$(jats__tidparse suite "$id")" ;;
        suite)      echo "jats://${nid%%//*}//" ;;
        id)         echo "$id" ;;
        params)     echo "$params" ;;
        relpath)    echo "${nid#*//}" ;;
        "")         warn "usage: jats__tidparse ELEM TESTID"
                    return 2 ;;
        *)          warn "unknown element: $elem"
                    return 2 ;;
    esac
}

jats__tlocate() {
    #
    # Locate test $1 under known paths
    #
    local t_id=$1
    local s_id
    local t_relpath
    local d=$'\t'
    s_id=$(jats__tidparse suite "$t_id")
    t_relpath=$(jats__tidparse relpath "$t_id")
    debug -v s_id t_relpath
    __jats__scan_known \
      | grep -Fe "${d}s_id=$s_id${d}" \
      | grep -Fe "${d}t_relpath=$t_relpath${d}" \
      | grep -Eoe "${d}s_dir=[^${d}]+${d}" \
      | cut -d"$d" -f2 \
      | cut -d= -f2- \
      | sed "s:$:/$t_relpath:" \
      | grep . \
     || return 3
}

jats__tprop() {
    #
    # Get test meta $1 about test at dir $2
    #
    local pname=$1
    local t_dir=$2
    local pvalue
    local s_dir
    local s_id
    local t_relpath
    case $pname in
        id|version) : ;;
        *)  warn "usage: jats__tprop id|version PATH"; return 2 ;;
    esac
    test -d "$t_dir" || {
        warn "no such test directory: $t_dir"
        return 3
    }
    s_dir=$(__jats__straverse "$t_dir")
    t_relpath=${t_dir#$s_dir/}
    case $pname in
        id)
            s_id=$(jats__sprop id "$s_dir")
            test -n "$s_id" || {
                s_id="jats://_jats_no_suite_//"
                warn "could not find suite id; made up one: s_id=$s_id"
            }
            echo "$s_id$t_relpath"
            ;;
        version)
            s_version=$(jats__sprop version "$s_dir")
            test -n "$s_version" || {
                s_version="0.0.0+_jats_no_version_"
                warn "could not find suite version; made up one: s_version=$s_version"
            }
            echo "$s_version"
            ;;
        *)
            warn "usage: jats__tprop id|version PATH"
            return 2
            ;;
    esac
}

__jats__straverse() {
    #
    # Traverse back from test dir $1 and print its suite dir
    #
    local tdir=$1
    local dir
    dir=$(readlink -f "$tdir")
    while true; do
        test -d "$dir/.jats" && echo "$dir" && return 0
        test "$dir" == "/" && return 1
        dir=${dir%/*}; test -n "$dir" || dir=/
    done
}

__jats__ckfmt() {
    #
    # Check if suite at $1 is compatible
    #
    local dir=$1
    local fmt
    fmt=$(jats__sprop format "$dir")
    test "$fmt" == "$JATS__COMPAT"
}

__jats__known() {
    #
    # List out known and existing paths
    #
    local trypath
    debug -v JATS__PATH
    for trypath in ${JATS__PATH//:/ }; do
        debug -v trypath
        test -n "$trypath" || continue
        test -d "$trypath" || continue
        echo "$trypath"
    done
}

__jats__scan_known() {
    #
    # Scan known paths for suites
    #
    local togo
    local s_dir
    local s_id
    local s_version
    local s_fmt
    local t_relpath
    local d=$'\t'
    arr__fromcmd togo __jats__known
    find "${togo[@]}" -type d -path '*/.jats' \
      | sed 's:/[.]jats$::' \
      | while read -r s_dir; do
            debug -v s_dir
            __jats__ckfmt "$s_dir" || continue
            s_id=$(jats__sprop id "$s_dir")
            s_version=$(jats__sprop version "$s_dir")
            find "$s_dir" -type f -name 'test' -printf '%P\n' \
              | sed 's:/test$::' \
              | while read -r t_relpath; do
                    echo "${d}s_dir=$s_dir${d}s_id=$s_id${d}s_version=$s_version${d}t_relpath=$t_relpath${d}"
                done
        done
}

#shellfu module-version=0.0.10
